;Http interface driver for lookout

(use-modules (ice-9 debug)

	     (srfi srfi-2)
             (alterator http session))

;;;;;;;;;;;;;;;;;;;;;;;; server engine
(define *server-socket* #f)
(define *socket-name* "/tmp/alterator-sock")

;process "get url" request
(define (http-cmd-get url)
  (format #t "getting url=~S~%" url)
  (format #t "current http-widgets is=~A~%" (from-session 'http-widgets))
  (and url
       (let ((dialogs (from-session 'http-dialogs)))
	 (and dialogs
	      (let ((dialog (cond-cdr (assq (sure-symbol url) (cdr dialogs)))))
		(if (not dialog)
		  (let ((dlg (init-popup (document:subdocument url))))
		    (and dlg
			 (set-cdr! dialogs (alist-set (sure-symbol url) dlg (cdr dialogs)))))
		  (dialog 'bubbling 'recreate)))))))

;invisible set for "event" request
(define (http-silent-process wid)
  (let ((widgets (cond-cdr (from-session 'http-widgets))))
    (and widgets
	 (and-let* ((name (sure-symbol (car wid)))
		    (widget (cond-cdr (assq name widgets)))
		    (attrs (cdr wid)))
		   (for-each (lambda (attr)
			       (let ((name (car attr))
				     (value (cdr attr)))
				 (widget 'silent-set name value)))
			     attrs)))))

;process "event emit" request
(define (http-cmd-event cmd)
  (let ((widgets (cond-cdr (from-session 'http-widgets))))
    (and widgets
	 (and-let* ((name (cond-cdr (command-arg-ref cmd 'name)))
		    (id   (cond-cdr (command-arg-ref cmd 'widget-id)))
		    (state (cond-cdr (command-arg-ref cmd 'state)))
		    (widget (cond-cdr (assq (sure-symbol id) widgets))))
		   ;first process state here
		   (for-each http-silent-process state)
		   ;then run callback (TODO: run in separate queue to support run-dialog like freeze
		   (widget 'call (sure-symbol name))))))

;destoy http application
(define (delete-http-application app)
  (close-port *server-socket*)
  (delete-file *socket-name*))


;create new socket
(define (make-http-application cmdline handler)
  (write-line "creating server side socket")
  (set! *server-socket* (socket PF_UNIX SOCK_STREAM 0))
  (bind *server-socket* AF_UNIX *socket-name*)
  (listen *server-socket* 10)
  (chmod *socket-name* #o666))

;process incoming loop requests
(define (process-http-request cmd)
  (format #t "req=~A~%" cmd)
  (let ((action (cond-cdr (command-arg-ref cmd 'action))))
    (and action
	 (case (sure-symbol action)
	   ((get) (write-line "this is a get action")
		  (http-cmd-get (cond-cdr (command-arg-ref cmd 'url))))
	   ((event) (write-line "this is a event notification")
		    (http-cmd-event cmd))
	   (else (error "unknown action" action))))))


;;;;;;;;;;;;;;;;;;;;;;;;; auth engine

;demo authorizarion
(define (auth-user user passwd)
  (format #t "authentication for user=~A, passwd=~A~%" user passwd)
  (and (string=? user "root")
       (string=? passwd "root")))

;auth processing here
(define (auth-http-request cmd next)
  (define (auth-permission-denied)
    `(auth-answer content ((command action "forbidden"))))

  (define (auth-permission-granted content . args)
    (and content (next content))
    `(auth-answer ,@args content ,(http-dump-queue)))
  
  (format #t "auth=~A~%>>>sessions=~S~%" cmd  http-session-list)
  (let ((content (cond-cdr (command-arg-ref cmd 'content)))
	(user	 (cond-cdr (command-arg-ref cmd 'user)))
	(password (cond-cdr (command-arg-ref cmd 'password)))
	(session-id (cond-cdr (command-arg-ref cmd 'session-id))))

    (cond
      (session-id 
	(let ((session (cond-cdr (assv (string->number session-id) http-session-list))))
	  (if (and session (string=? (cond-cdr (assq 'http-user session)) user))
	    (with-fluids ((http-session session))
			 (auth-permission-granted content))
	    (auth-permission-denied))))
      ((and user password (auth-user user password))
       (let ((session (make-http-session user))
	     (session-id (make-session-id)))
	 (set! http-session-list (alist-set session-id session http-session-list))
	 (with-fluids ((http-session session))
		      (auth-permission-granted content
			'user user
			'session-id session-id))))
      (else (auth-permission-denied)))))

  
;listen on socket and run appropriate processing
(define (exec-http-application)
  (write-line "listen for connections ...")
  (let listen-loop ((c (car (accept *server-socket*))))
    (write-line "accepting connection and process commands")
    (let command-loop ((cmd (read c)))
      (and (not (eof-object? cmd))
	   (begin (write (auth-http-request cmd process-http-request) c)
		  (command-loop (read c)))))
    (write-line "go to next processing")
    (listen-loop (car (accept *server-socket*)))))


;;;;;;;;;;;;;;;;;;;;;;;;;; widget engines

(define (make-http-widget-dispatcher)
  (let ((attributes '()))
    (lambda (id command . args)
      (case command
	((set silent-set) (let ((name (sure-symbol (car args)))
				(value (cadr args)))
			    (set! attributes (alist-set name value
							attributes))
			    (and (eq? command 'set)
				 (not (eq? name 'widget-id)) 
				 (http-post-event 'action "set"
						  'widget-id id
						  'attr name
						  'value value))))
	((get) (cond-cdr (assq (sure-symbol (car args)) attributes)))
	((bubbling) (case (car args)
		      ((recreate)
		       (for-each (lambda (attr)
				   (let ((name (car attr))
					 (value (cdr attr)))
				     (and (not (eq? name 'widget-id))
					  (http-post-event 'action "set"
							   'widget-id id
							   'attr name
							   'value value))))
				 attributes))))
	(else
	  (error "unknown operation for http widget"))))))

(define (make-http-widget type . rest)
  (let ((dispatcher (if (null? rest) (make-http-widget-dispatcher) (car rest))))
    (lambda (events-holder super)
      (let ((id #f)
	    (parent #f)
	    (events events-holder) ;events dispatcher
	    (pending-requests (list #f)))
      (lambda (command . args)
	(case command
	  ((get-http-id) id)
	  ((new) (and (not id)
		      (begin
			(set! id ((cdr (from-session 'http-id-generator))))
			(set! parent (if (procedure? (car args))
				       ((car args) 'get-http-id)
				       ""))
			(let ((widgets (from-session 'http-widgets)))
			  (set-cdr! widgets (alist-set (sure-symbol id) (car super) (cdr widgets))))
			(http-post-event 'action "new"
					 'widget-id id
					 'type type
					 'parent parent))))
	  ((post-new)
	         ;register available events
		 (for-each (lambda (x)
			     (http-post-event 'action "create-event"
					      'widget-id id
					      'value x))
			   (events 'list))
	         ;process pending requests
		 (for-each (lambda (x)
		   		(apply dispatcher widget (car x) (cdr x))) ; apply delayed requests
			     (cdr pending-requests)))
	  ((bubbling) (case (car args)
			((recreate)
			 (and id
			      (begin
				(http-post-event 'action "new"
						 'widget-id id
						 'type type
						 'parent parent)
				(apply dispatcher id command args)
				(for-each (lambda (x)
					    (http-post-event 'action "create-event"
							     'widget-id id
							     'value x))
					  (events 'list)))))))

	  (else  (if id
		   (apply dispatcher id command args)
		   (begin (append! pending-requests (list (cons command args)))
			  #f)))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;driver description
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(declare-widget button (make-http-widget "button") text)
(declare-widget edit (make-http-widget "edit") text)
(declare-widget hbox (make-http-widget "hbox"))
(declare-widget vbox (make-http-widget "vbox"))

;wrapper over tab-panel
;insert widgets and generate also tab-panel attribute
(define (tab tab-id tab-label . widgets)
  (let ((vb (id tab-id (apply vbox widgets))))
   `(,vb ,(tab-panel tab-label (cadr vb)))))

(declare-application-constructor make-http-application)
(declare-application-destructor delete-http-application)
(declare-application-runner exec-http-application)

(declare-root-widget (make-http-widget "vbox"))
(declare-popup-widget (make-http-widget "vbox"))

