#!/usr/bin/guile -s
!#
(use-modules (ice-9 debug)
             (srfi srfi-1)
	     (srfi srfi-13)

	     (alterator command)
	     (alterator str)
	     (alterator algo)
	     (alterator http cgi)
	     (alterator http request))

(define *socket-name* "/tmp/alterator-sock")

;get cookies and envelop our message for auth module
;direct auth messages pass as is
(define (auth-pre-process message)
  (if (eq? (car message) 'auth-request)
    message
    (let* ((cookie (get-cookie))
	   (user (cond-cdr (assoc "alterator-user" cookie)))
	   (password (cond-cdr (assoc "alterator-password" cookie)))
	   (session-id (cond-cdr (assoc "alterator-session-id" cookie)))
	   (remote-addr (remote-address)))
	   
      (append '(auth-request)
	      (concatenate
		(filter and
			(list
			  (and remote-addr `(remote-host ,remote-addr))
			  (and user `(user ,user))
			  (and password `(password ,password))
			  (and session-id `(session-id ,session-id)))))
	      (list 'content message)))))

;set cookies to message if auth module want's this
;extract enveloped message
(define (auth-post-process message)
  (if (eof-object? message)
    '()
    (begin 
      (let* ((content (cond-cdr (command-arg-ref message 'content)))
	     (user    (cond-cdr (command-arg-ref message 'user)))
	     (session-id (cond-cdr (command-arg-ref message 'session-id))))
	;setup cookies
	(and user
	     session-id
	     (set-cookie `((alterator-user . ,user)
			   (alterator-session-id . ,session-id))))
	;return content
	content))))

;communicate with server
(define (alterator<->cgi message)
  (let ((s (socket PF_UNIX SOCK_STREAM 0)))
    (connect s AF_UNIX *socket-name*)
    (write (auth-pre-process message) s)
    (begin-1
      (auth-post-process (read s))
      (close-port s))))

(define (write-content inputreq)
  (let ((message (with-input-from-string inputreq (lambda() (read)))))
    (and (not (eof-object? message))
	 	   (put-message
		     (with-output-to-string
		       (lambda()
			 (let ((answer (alterator<->cgi message)))
			   (write-text-xml)
			   (write-line "<alterator-answer>")
			   (map (lambda (cmd)
				  (display " ")(write-request cmd)(newline))
				answer)
			   (write-line "</alterator-answer>"))))))))

(get-message
  (lambda (msg)
    (if (string=? (request-method)  "GET")
      (write-unsupported)
      (write-content msg))))

