;main engine to lookout incput documents
(define-module (alterator lookout document)
	       :use-module (alterator algo)
	       :use-module (alterator command)
	       :use-module (alterator lookout uri)
	       :use-module (alterator lookout context)
	       :use-module (alterator lookout globals)
	       :use-module (alterator lookout mapper)
	       :use-module (alterator lookout widgets)
	       :use-module (alterator lookout goto)
	       :use-module (alterator lookout ids)
	       :use-module (alterator lookout popup)
               :export (declare-root-widget
	       		lookout-root-widget
			
	       		document:preprocess
			document:subdocument

			preprocess-file
			init-popup

			document:popup
			document:end))

;;;;;;;;;;;;;;;;;;;;
(define lookout-root-widget #f)

(define-macro (declare-root-widget func)
	      `(set! lookout-root-widget ,func))

;;;;;;;;;;;;;;;;;;;; main preprocessing

;wrapper over command-arg-ref to made code simpler
(define (get-param cmd param)
  (cond-cdr (command-arg-ref cmd param)))

;analyze link instruction
(define (process-link port instruction initial)
  (or
    (let ((type (get-param instruction 'type)))
      (case type
	((popup)
	 (let ((text (get-param instruction 'text))
	       (path (get-param instruction 'to)))
	   (and text path
		(append initial
			(list `(button ,text
				       (on-click
					 (document:popup ,@path))))))))
	((rollup)
	 (let ((id (gensym))
	       (text (get-param instruction 'text))
	       (path (get-param instruction 'to)))
	   (and text path
		(append initial
			(list `(button ,text
				       (on-click (,id hidden (not (,id hidden)))))
			      `(id ',id (vbox hidden #t  ,(apply document:subdocument path))))))))
	((insert subdocument)
	 (let ((path (get-param instruction 'to)))
	   (and path
		(append initial
			(if (eq? type 'insert)
			  (document:preprocess (car path))
			  (list (apply document:subdocument path)))))))
	(else (error "unknown link type " type))))
    initial))


;run subroutine to process next enviroment
(define (process-enviroment port tag initial)
    (append  initial
	     (list `(,@(cdr tag)
		      ,@(read-enviroment port (cadr tag) '() )))))

;processing surround instruction
(define (process-surround port instruction initial)
    (append initial
	    ;append result of preprocessing of other file with rest: our rest file
	    (apply document:preprocess
		    (cadr instruction)
		    (list (cons 'document:quote (read-enviroment port (gensym) '() ))))))


;analyze pair instruction
(define (process-pair port instruction initial)
  (case (car instruction)
    ((document:link) (process-link port instruction initial))
    ((document:envelop) (process-enviroment port instruction initial))
    ((document:end-envelop) (error "parse-error, unexpected end-envelop" (cadr instruction)))
    ((document:quote) (append initial (cdr instruction)))
    ((document:surround) (process-surround port instruction initial))
    (else (append initial (list instruction)))))

;analyze instruction we read
(define (process-instruction port instruction initial)
  (cond 
    ((pair? instruction) (process-pair port instruction initial))
    (else (append initial (list instruction)))))

;get next intruction from rest
;demo: return all the rest
(define (get-from-rest port)
  (let ((rest (cadr port)))
    (if (null? rest)
      (read (car port)) ;eof here
      (begin-1 (car rest)
	       (set-car! (cdr port) (cdr rest))))))


;get next instruction from port
(define (get-from-port port)
  (read (car port)))

;get next instruction from input stream (port or already readed data)
;first try to read from file
;then try to return rest
;otherwise return eof
(define (get-instruction port)
;  (format #t "port content is =~A~%" port)
  (let ((result (get-from-port port)))
    (if (eof-object? result) ;first try to read from file
      (get-from-rest port)
      result)));return next result from file

;go to the depth of instruction and parse it
(define (parse-in-depth instruction)
;  (format #t "parsing instruction=~A~%" instruction)
  (if (eof-object? instruction)
       instruction
       (cond
	 ((and (pair? instruction)
	       (eq? (car instruction) 'document:quote))
	  instruction)        ;don't parse quotes
	 ((list? instruction)
	  (apply preprocess-file "/dev/null" instruction))
	 (else instruction)))) ; non-lists are not interest for us

;collect data from some enviroment
;enviroment ends with end-of-file or at document:ent-evelop instruction
(define (read-enviroment port tagname initial)
  (let ((instruction (parse-in-depth (get-instruction port))))
    (if (or (eof-object? instruction)
	    (and (pair? instruction)
		 (eq? (car instruction) 'document:end-envelop)
		 (eq? (cadr instruction) tagname)))
	    initial
	    (read-enviroment port tagname
			     (process-instruction port instruction initial)))))

;preprocess dialog with links and eviroments
(define (preprocess-file file . rest)
  (call-with-input-file file
			(lambda (port)
			  (read-enviroment (append (list port) (list rest))
					   (gensym) '() ))))

;same as preprocess-file, but also resolving url
(define (document:preprocess  path . rest)
  (let ((path (lookout-convert-path path)))
    (or  (and path (apply preprocess-file (car path) rest))
	 '())))


;create new  all-sufficient document
;todo: will be definition postprocessing here
(define (document:subdocument . url)
  (let ((path (lookout-convert-uri url))
	(root (make-widget lookout-root-widget)))
    (and path
	 (list 'widget
	       root
	       (lambda (parent)
		 (root 'new parent)
		 (with-goto-frame root
				(with-ids (make-ids-holder)
					(with-globals (extend-globals '())
						(command-for-each (lambda(name value)
								    (set-global! name value))
								  path)
						(widget-parse-args root
								   (map 
								     (lambda (x)
								       (eval x (current-module)))
									(preprocess-file (car path))))
						(root 'set-context (fluid-ref lookout-context)))))
		 root)))))

(define (init-popup main-widget)
  (and (pair? main-widget)
       (make-popup lookout-popup-widget (from-context 'popup) main-widget)))


;resolve uri, and run new dialog
;return dialog's return code
(define (document:popup . url)
  (let ((dlg (init-popup (apply document:subdocument url))))
    (and (procedure? dlg)
	 (with-popup dlg
		      (begin-1
			(dlg 'start)
			(dlg 'delete))))))

;exit current dialog modal loop
(define (document:end . args)
  (apply (from-context 'popup) 'stop args))
 
