;description: functions to change content of some widgets
;goto-in-widget replace content of selected widget
;goto function same as goto-in-widget, but works in some default goto-frame
;goto-frame lives in current lookout context

(define-module (alterator lookout goto)
	       :use-module (ice-9 rdelim)
	       :use-module (ice-9 syncase)
	       :use-module (alterator algo)
	       :use-module (alterator lookout context)
	       :use-module (alterator lookout globals)
	       :use-module (alterator lookout uri)
	       :use-module (alterator lookout mapper)
	       :use-module (alterator lookout widgets)
	       :use-module (alterator lookout events)
	       :use-module (alterator lookout document)
	       :export (document:replace
			 document:replace-in-widget
			 
			 on-goto
			 on-before-goto
			 
			 with-goto-frame))

(declare-event on-goto url)
(declare-event on-before-goto url)

;replace content in some widget
(define (document:replace-in-widget widget url . args)
  (let ((path (lookout-convert-uri (cons url args))))
    ;work with widget's context and it's globals
    (with-context (widget 'get-context)
		  (and path
		       (begin (widget 'delete-subwidgets)
			      (reset-globals! (cdr path))
			      (widget-parse-args
				widget
				(map (lambda (x) (eval x (current-module))) 
				     (preprocess-file (car path))))
			      (widget 'show-subwidgets))))))

(define (document:replace url . args)
  (let ((current-goto-frame (from-context 'goto-frame)))
    (current-goto-frame 'call on-before-goto url)
    (apply document:replace-in-widget current-goto-frame url args)
    (current-goto-frame 'call on-goto url)))

(define-syntax with-goto-frame
  (syntax-rules()
    ((_ frame instruction instructions ...)
     (with-changed-context 'goto-frame frame instruction instructions ...))))


